/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2005 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.2
 *
 ******************************************************************************/

/* Size of stack-allocated buffer to use for nxo_file_readline().  If this
 * overflows, heap allocation is used. */
#ifdef CW_DBG
#define CW_NXO_FILE_READLINE_BUFSIZE 25
#else
#define CW_NXO_FILE_READLINE_BUFSIZE 100
#endif

typedef struct cw_nxoe_file_s cw_nxoe_file_t;

struct cw_nxoe_file_s
{
    cw_nxoe_t nxoe;
#ifdef CW_THREADS
    /* Access is locked if this object has the locking bit set. */
    cw_mtx_t lock;
#endif

    char *origin;
    uint32_t olen;

    enum
    {
	FILE_NONE,
#ifdef CW_POSIX_FILE
	FILE_POSIX,
#endif
	FILE_SYNTHETIC
    } mode:2;
    bool nonblocking:1;

    union
    {
	struct
	{
	    cw_nxo_file_read_t *read_f;
	    cw_nxo_file_write_t *write_f;
	    cw_nxo_file_ref_iter_t *ref_iter_f;
	    cw_nxo_file_delete_t *delete_f;
	    void *arg;
	    cw_nxoi_t position;
	} s;
#ifdef CW_POSIX_FILE
	struct
	{
	    int32_t fd;
	    bool close;
	} p;
#endif
    } f;
    /* Buffering. */
    char *buffer;
    uint32_t buffer_size;
    enum
    {
	BUFFER_EMPTY,
	BUFFER_READ,
	BUFFER_WRITE
    } buffer_mode;
    uint32_t buffer_offset;
};

/* Private, but needed for the inlining of nxoe_l_file_delete(). */
cw_nxn_t
nxo_p_file_buffer_flush(cw_nxoe_file_t *a_file);

#ifndef CW_USE_INLINES
bool
nxoe_l_file_delete(cw_nxoe_t *a_nxoe, uint32_t a_iter);

cw_nxoe_t *
nxoe_l_file_ref_iter(cw_nxoe_t *a_nxoe, bool a_reset);
#endif

#if (defined(CW_USE_INLINES) || defined(CW_NXO_FILE_C_))
CW_INLINE bool
nxoe_l_file_delete(cw_nxoe_t *a_nxoe, uint32_t a_iter)
{
    cw_nxoe_file_t *file;

    file = (cw_nxoe_file_t *) a_nxoe;

    cw_check_ptr(file);
    cw_dassert(file->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(file->nxoe.type == NXOT_FILE);

    nxo_p_file_buffer_flush(file);
    if (file->buffer != NULL)
    {
	nxa_free(file->buffer, file->buffer_size);
    }
#ifdef CW_THREADS
    if (file->nxoe.locking)
    {
	mtx_delete(&file->lock);
    }
#endif
    if (file->origin != NULL)
    {
	nxa_free(file->origin, file->olen);
    }
    switch (file->mode)
    {
	case FILE_NONE:
	{
	    break;
	}
#ifdef CW_POSIX_FILE
	case FILE_POSIX:
	{
	    if (file->f.p.close)
	    {
		close(file->f.p.fd);
	    }
	    break;
	}
#endif
	case FILE_SYNTHETIC:
	{
	    if (file->f.s.delete_f != NULL)
	    {
		file->f.s.delete_f(file->f.s.arg);
	    }
	    break;
	}
    }

    nxa_free(file, sizeof(cw_nxoe_file_t));

    return false;
}

CW_INLINE cw_nxoe_t *
nxoe_l_file_ref_iter(cw_nxoe_t *a_nxoe, bool a_reset)
{
    cw_nxoe_t *retval;
    cw_nxoe_file_t *file;

    file = (cw_nxoe_file_t *) a_nxoe;

    if (file->mode == FILE_SYNTHETIC && file->f.s.ref_iter_f != NULL)
    {
	retval = file->f.s.ref_iter_f(file->f.s.arg, a_reset);
    }
    else
    {
	retval = NULL;
    }

    return retval;
}
#endif /* (defined(CW_USE_INLINES) || defined(CW_NXO_FILE_C_)) */
